/*
 *	LCD interface example
 *	Uses routines from delay.c
 *	This code will interface to a standard LCD controller
 *	like the Hitachi HD44780. It uses it in 4 or 8 bit mode
 *	
 */

#include	<pic18.h>
#include	"LCD.H"
#include	"Delay.h"

static bit fourbit;		// four or eight bit mode?
unsigned char ENSelect;

#ifdef CHECKBUSY

unsigned char 
lcd_read_cmd_nowait(void)
{
	unsigned char c, readc;

	LCD_DATA_TRIS	 = INPUT_DATA;

	LCD_RW = 1; // Read LCD
	asm("nop"); // short propagation delay
	asm("nop"); // short propagation delay
	DelayUs(15);

	if (fourbit)
	{
		LCD_STROBE_READ(readc); // Read high nibble
		// Move 4 bits to high nibble while zeroing low nibble
		c = ( ( readc << 4 ) & 0xF0 ); 
		LCD_STROBE_READ(readc); // Read low nibble
    		c |= ( readc & 0x0F ); // Or in 4 more bits to low nibble
	}
	else
	{
		LCD_STROBE_READ(readc); 
		c = readc;
	}
	LCD_RW = 0; // Return to default mode of writing LCD
	LCD_DATA_TRIS	 = OUTPUT_DATA; // Return to default mode of writing LCD

	return(c);
}

void
lcd_check_busy(void) // Return when the LCD is no longer busy, or we've waiting long enough!
{
	// To avoid hanging forever in event there's a bad or 
	// missing LCD on hardware.  Will just run SLOW, but still run.
	unsigned int retry; 
	unsigned char c;

	for (retry=1000; retry-- > 0; ) {
		c = lcd_read_cmd_nowait();
		if (0==(c&0x80)) break; // Check busy bit.  If zero, no longer busy
	}
}

#endif

/* send a command to the LCD */
void
lcd_cmd(unsigned char c)
{
	LCD_WAIT; // may check LCD busy flag, or just delay a little, depending on lcd.h

	if (fourbit)
	{
		LCD_DATA = ( ( c >> 4 ) & 0x0F );
		LCD_STROBE();
    		LCD_DATA = ( c & 0x0F | 0x80);
		LCD_STROBE();
	}
	else
	{
		LCD_DATA = c;
		LCD_STROBE();
	}
}

unsigned char
lcd_check_displ(void)
{
	// To avoid hanging forever in event there's a bad or 
	// missing LCD on hardware.  Will just run SLOW, but still run.
	unsigned int retry; 
	unsigned char c;

	for (retry=1000; retry-- > 0; ) {
		c = lcd_read_cmd_nowait();
		if (0==(c&0x80)) 
		{
			return c&0x7F; // Check busy bit.  If zero, no longer busy
		}
	}
	return 0;
}

/* send data to the LCD */
void
lcd_data(unsigned char c)
{
	LCD_WAIT; // may check LCD busy flag, or just delay a little, depending on lcd.h

	LCD_DATA = 0;
	LCD_RS = 1;
	if (fourbit)
	{
    		LCD_DATA |= ( ( c >> 4 ) & 0x0F );      
		LCD_STROBE();
		LCD_DATA &= 0xF0;
		LCD_DATA |= ( c & 0x0F | 0x80); 
		LCD_STROBE();
	}
	else
	{
		LCD_DATA = c;
		LCD_STROBE();
	}
	LCD_RS = 0;
  
}

void
lcd_data_q(unsigned char c)
{
	LCD_DATA = 0;
	LCD_RS = 1;
	if (fourbit)
	{
    		LCD_DATA |= ( ( c >> 4 ) & 0x0F );      
		LCD_STROBE();
		LCD_DATA &= 0xF0;
		LCD_DATA |= ( c & 0x0F | 0x80); 
		LCD_STROBE();
	}
	else
	{
		LCD_DATA = c;
		LCD_STROBE();
	}
	LCD_RS = 0;
}

/* write a string of chars to the LCD */

void
lcd_puts(const char * s)
{
	while(*s)
		lcd_data(*s++);
}

void lcd_putsLn(const char * s)
{
  while(*s)
  {
    if(lcd_check_displ()==LineLength)
    {
      lcd_cmd(0xC0);
      LCD_WAIT;
    }
    lcd_data_q(*s++);
  }
}

void lcd_putnum(unsigned char num)
{
  if(num>99) lcd_data(num/100 + '0');
  if(num>9) lcd_data((num/10)%10 + '0');
  lcd_data(num%10 + '0');
}

void lcd_putuint(unsigned int num)
{
  if(num>9999) lcd_data(num/10000 + '0');
  if(num>999) lcd_data((num/1000)%10 + '0');
  if(num>99) lcd_data((num/100)%10 + '0');
  if(num>9) lcd_data((num/10)%10 + '0');
  lcd_data(num%10 + '0');
}

void lcd_puthex(unsigned char num)
{
  unsigned char i;
  i=num>>4;
  if(i<10) lcd_putch('0'+i); else lcd_putch('A'-10+i);
  i=num&0x0F;
  if(i<10) lcd_putch('0'+i); else lcd_putch('A'-10+i);
}

void lcd_gotoLn(unsigned char pos)
{
  #if(DisplayType==LCDTwoEN)
    if(pos&0x80) ENSelect=2; else ENSelect=1;
    pos&=0x7F;
  #endif
  if(pos<0x40) lcd_cmd(0x80|pos);
          else lcd_cmd(0xC0|pos);
}

void lcd_gotoXY(unsigned char Radek,unsigned char Sloupec)
{
  switch(Radek)
  {
    case 0:lcd_gotoLn(LcdLn1+Sloupec);return;
    case 1:lcd_gotoLn(LcdLn2+Sloupec);return;
    case 2:lcd_gotoLn(LcdLn3+Sloupec);return;
    case 3:lcd_gotoLn(LcdLn4+Sloupec);return;
  }
}

void lcd_LoadCGRAM(unsigned char Pozice,const unsigned char * Znak)
{
  lcd_cmd(0x40 + Pozice);
  Pozice=8;
  while(Pozice--) lcd_data(*Znak++);
}

/* initialize the LCD */
void
lcd_init(unsigned char mode)
{
	char init_value;
  ENSelect=3;

	fourbit		= 0;
	if (mode == FOURBIT_MODE){
		fourbit = 1;
		init_value = 0x3;
	}else{
		init_value = 0x3F;
	}
	LCD_RS = 0;
  #if(DisplayType==LCDOneEN)
	  LCD_EN = 0;
  #endif
  #if(DisplayType==LCDTwoEN)
	  LCD_EN1 = 0;
	  LCD_EN2 = 0;
  #endif
	LCD_RW = 0;
	LCD_RS_TRIS	 = OUTPUT_PIN;
  #if(DisplayType==LCDOneEN)
	  LCD_EN_TRIS = OUTPUT_PIN;
  #endif
  #if(DisplayType==LCDTwoEN)
	  LCD_EN1_TRIS = OUTPUT_PIN;
	  LCD_EN2_TRIS = OUTPUT_PIN;
  #endif
	LCD_RW_TRIS	 = OUTPUT_PIN;
	LCD_DATA_TRIS	 = OUTPUT_DATA;
	DelayMs(15);
	LCD_DATA	 = init_value;
	LCD_STROBE();
	DelayMs(5);
	LCD_DATA	 = init_value;
	LCD_STROBE();
	DelayUs(200);
	LCD_DATA	 = init_value;
	LCD_STROBE();
	
	if (fourbit){
		LCD_WAIT; //may check LCD busy flag, or just delay a little, depending on lcd.h
		LCD_DATA = 0x2; // Set 4-bit mode
		LCD_STROBE();

		lcd_cmd(0x28); // Function Set
	}else{
		lcd_cmd(0x38);
	}
	lcd_cmd(0xC); //Display On, Cursor Off, Cursor Blink off
	lcd_cmd(0x1); //Display Clear
	lcd_cmd(0x6); //Entry Mode
	lcd_cmd(0x80); //Initialize DDRAM address to zero
  ENSelect=0;
}
